/**
 * Project Salmon (Habbo Hotel r63b Emulator)
 *
 * Copyright (c) 2013 - 2014, Jacob Pollack.
 **
 * SVN: svn://jamach.servehttp.com/projectSalmon
 **
 */

#include "net/connectionreactor.h"

#include <cstring> // memcpy
#include <errno.h>

#ifdef DEBUG
#include <iostream>
#endif

#include <sys/socket.h>
#include <sys/time.h>

#include "exceptions/critexception.h"
#include "net/session.h"

using std::map;
using std::pair;

#ifdef DEBUG
using std::cout;
using std::endl;
#endif

// See header file.
ConnectionReactor::ConnectionReactor(const int serverFD) : serverFD(serverFD) {
  this->connections = map<int, Session *>();
}

// See header file.
ConnectionReactor::~ConnectionReactor() {
  // Shuts down all of the active connections.
  while (false == this->connections.empty()) {
    this->removeConnection(this->connections.begin()->second);
  }
}

// See header file.
void ConnectionReactor::handleConnections() {
  bool shutdown = false;
  int result,        // Stores the result from various function calls.
      maxFD,         // Stores the maximum file descriptor.
      numUpdatedFDs; // Stores the number of updated file descriptors.
  fd_set master, worker;
  Session *client;

  // Zero-fills the master and worker sets.
  FD_ZERO(&master);
  FD_ZERO(&worker);

  // Puts the server socket's file descriptor in the master set.
  FD_SET(this->serverFD, &master);

  // Updates the maximum file descriptor.
  maxFD = this->serverFD;

  // Iterates until we are prompted to shutdown.
  while (false == shutdown) {
    // Copy the master set to the worker set.
    memcpy(&worker, &master, sizeof(master));

    #ifdef DEBUG
    cout << "--> Blocked, waiting to handle one or more events." << endl;
    #endif

    // Blocks until a socket within 0 ... maxFD + 1 has updated its state.
    result = select(maxFD + 1, &worker, NULL, NULL, NULL);

    #ifdef DEBUG
    cout << "--> Unblocked, handling " << result << " events(s)." << endl;
    #endif

    // Determines if there was an error on the call to select.
    if (-1 == result) {
      throw CritException("One or more errors occurred when making a call to select(). Use gdb.", __LINE__, __FILE__);
    }

    // Updates the number of updated file descriptors.
    numUpdatedFDs = result;

    // Determines if the server socket's file descriptor has updated its state.
    if (false != FD_ISSET(this->serverFD, &worker)) {
      bool done = false;

      // Iterates over all of the incoming connection requests.
      while (false == done) {
        // Accepts an incoming connection.
        result = accept(this->serverFD, NULL, NULL);

        // Determines if an error occurred accepting the incoming 
        // connection. If so, ignore it and continue. If there are no 
        // connections to accept, update the loop invariant. Otherwise, 
        // accept the new connection.
        if (-1 == result) {
          done = (EWOULDBLOCK == errno);
        } else {
          #ifdef DEBUG
          cout << "--> Accepted incoming connection." << endl;
          #endif

          // Adds the connection to the connections list.
          this->addConnection(new Session(result));

          // Adds the connection to the master set.
          FD_SET(result, &master);

          // Determines if we need to update the maximum file descriptor. If so, 
          // update it.
          if (result > maxFD) {
            maxFD = result;
          }
        }
      }

      // Updates the number of updated file descriptors.
      numUpdatedFDs -= 1;
    }

    // Iterates over all of the active connections and handles the incoming 
    // requests.
    // :TODO: Take another look at the following loop; maybe another, easier 
    //        way to structure it (choose a better loop invariant?).
    map<int, Session *>::iterator it = this->connections.begin();

    while ((it != this->connections.end()) && (0 < numUpdatedFDs)) {
      // Determines if the current socket's file descriptor has updated its 
      // state.
      if (false != FD_ISSET(it->first, &worker)) {
        #ifdef DEBUG
        cout << "--> Handling existing connection." << endl;
        #endif

        // Determines if the current connection is alive. If so, handle the 
        // incoming request, otherwise drop the connection.
        if (true == it->second->isAlive()) {
          // :TODO: ** LEFT OFF HERE **
        } else {
          // Determines if we need to update the maximum file descriptor. If so, 
          // update it.
          if (maxFD == it->first) {
            // Iterates over all of the file descriptors in the master set and 
            // decrements the maxFD respectively.
            while (false == FD_ISSET(maxFD, &master)) {
              maxFD -= 1;
            }
          }

          // Removes the connection from the master set.
          FD_CLR(it->first, &master);

          // Backs up the current client. This is to avoid a seg fault on the 
          // following lines.
          client = it->second;

          // Update the loop invariant and decrement the number of updated FDs.
          ++it;
          numUpdatedFDs -= 1;

          // Removes the connection from the connections list.
          this->removeConnection(client);

          continue;
        }

        // Updates the number of updated file descriptors.
        numUpdatedFDs -= 1;
      }

      // Increments the iterator.
      ++it;
    }
  }
}

// See header file.
void ConnectionReactor::addConnection(Session * const session) {
  this->connections.insert(pair<int, Session*>(session->getFileDes(), session));

  #ifdef DEBUG
  cout << "--> New session (on socket " << session->getFileDes() << ") connected!" << endl;
  #endif
}

// See header file.
void ConnectionReactor::removeConnection(Session * const session) {
  // Removes the session from the list of connections.
  const int result = this->connections.erase(session->getFileDes());

  // Determines if we are removing a connection that does not exist. If so, 
  // log the critical exception.
  if (0 == result) {
    throw CritException("Attempted to remove a connection that does not exist.", __LINE__, __FILE__);
  }

  #ifdef DEBUG
  cout << "--> Session (on socket " << session->getFileDes() << ") disconnected!" << endl;
  #endif

  delete session;
}